> module Yahc.Command where
> import Yahc.Checker
> import Control.Arrow
> import Data.Char (isSpace, toUpper)
> import Control.Monad

> data CmdName = Quit
>              | Help
>              | Show
>              | Start
>              | AddRule
>              | Next
>              | Undo
>                deriving (Eq, Show)

> data Command = Command { cmdname :: CmdName
>                        , action :: Action
>                        , longStr :: String
>                        , short :: String
>                        , help  :: String
>                        } 
> helpShow = unlines [
>              "show [rules | goal | derivation]"
>            , "muestran las reglas existentes,"
>            , "         el objetivo,"
>            , "         y la derivación hasta ahora,"
>            , "respectivamente."
>            ]

> helpUndo = unlines [
>              "undo"
>            , "deshace la última regla aplicada"
>            ]

> helpStart = unlines [
>              "start formula formula'"
>            , "comienza una nueva derivación y descarta"
>            , "la derivación actual."
>            ]


> helpAddRule = unlines [
>                "add nombre abreviatura"
>               , "agrega la última derivación como una nueva regla."
>               ]


> helpNext = unlines [
>              "next regla resultado"
>            , "trata de aplicar la regla "
>            , "a la conclusión de la derivación actual"
>            , "si tiene éxito agrega este paso a la "
>            , "derivación; si no tiene éxito no hace nada."
>            ]


> helpHelp = unlines [
>            "Los comandos existentes son (entre paréntesis las versiones cortas)"
>            , "  show  (sh)"
>            , "  help  (h)"
>            , "  start (s)"
>            , "  next  (n)"
>            , "  quit  (q)"
>            , "  add   (a)"
>            , "Puedes obtener más información sobre cada comando"
>            , "haciendo \"help comando\""
>            ]

> readCmd :: [Command] -> String -> Maybe Command
> readCmd cmds = uncurry mplus <<< lookupCmd short &&& lookupCmd longStr
>     where lookupCmd on what = lookup what . map (on &&& id) $ cmds

> parseCommand :: [Command] -> String -> Maybe (Action, String)
> parseCommand cmds line = lookupL line >>= \act -> return (action act,rest)
>     where (cmd, rest) = (readCmd cmds *** tail') . break isSpace $ line
>           lookupL = return cmd
>           tail' xs = if null xs then xs else tail xs
