
> module Yahc.Subst where
> import Prelude hiding (lookup)
> import Yahc.Expr
> import Yahc.Parser
> import Yahc.Rules
> import Yahc.Utils
> import Data.List (intersect, union, nub)
> import Yahc.Missing
> import Control.Monad
> import Control.Arrow
> import qualified Yahc.Unicode as U

TODO: Deberíamos cambiarlo por Map!

> class Substituible a where
>     mkvar :: VarName -> a
>     (-/-) :: a -> Subst a -> a

> type Subst a = [(VarName,a)]

> lookup :: (Eq a, Substituible a) => Subst a -> VarName -> a
> lookup [] x = mkvar x
> lookup ((u,e):s) x = if u == x then e else lookup s x

> instance Substituible QExpr where
>     mkvar = QVar
>     (QVar x) -/- sigma = lookup sigma x
>     (QCon k) -/- sigma = con k
>     (QUnSym s e) -/- sigma = QUnSym s (e -/- sigma)
>     (QSym s es) -/- sigma = QSym s $ map (-/- sigma) es

> instance Substituible Expr where
>     mkvar = Var
>     e -/- sigma = unlift (lift e -/- map (\(v,e') -> (v,lift e')) sigma)

> merge      :: Subst QExpr -> Subst QExpr -> [Subst QExpr]
> merge s1 s2 = if agree then return (s1 `union` s2) else []
>  where agree = all (\v -> (var v) -/- s1 == (var v) -/- s2) (map fst s1 `intersect` map fst s2)

> merge' :: [Subst QExpr] -> [Subst QExpr] -> [Subst QExpr]
> merge' ss ss' = concat [merge s s' | s <- ss, s' <- ss']

> match :: QExpr -> QExpr -> [Subst QExpr]
> match (QVar v) e = return [(v,e)]
> match (QCon k) (QCon k') | k == k'   = return []
>                          | otherwise = []
> match (QUnSym s e) (QUnSym s' e') | s == s' = match e e'
>                                   | otherwise = [] 

> match (QSym s es) (QSym s' es') | s==s'     = matches s (terms es) (terms es')
>                                 | otherwise = []
> match _ _ = []

> terms :: [QExpr] -> [([QExpr],[QExpr])]
> terms es = [splitAt i es | i <- [1..length es-1]]

> matches :: Symbol -> [([QExpr],[QExpr])] -> [([QExpr],[QExpr])] -> [Subst QExpr]
> matches s ps es | not (isAC s) = concat [merge' (matchReb (p,e)) (matchReb (p',e'))
>                                        | (p,p') <- ps, (e,e') <- es] 
>                 | otherwise  = concat [merge' (matchReb (p,e)) (matchReb (p',e'))
>                                        ++ merge' (matchReb (p',e)) (matchReb (p,e'))
>                                        | (p,p') <- ps
>                                        , (e,e') <- es]
>     where reb        :: [QExpr] -> QExpr
>           reb (e:[]) = e
>           reb es     = QSym s es
>           matchReb :: ([QExpr],[QExpr]) -> [Subst QExpr]
>           matchReb = uncurry match . (reb *** reb)


> type Location = ([Int],[Int])

> subqexprs :: QExpr -> [(Location,QExpr)]
> subqexprs e@(QVar v) = [(([],[]),e)]
> subqexprs e@(QCon k) = [(([],[]),e)]
> subqexprs e@(QUnSym s e') = (([],[]),e): map (((0:) *** id) *** id) (subqexprs e')
> subqexprs e@(QSym s fs) = (([],[]),e) : (map (((,) []) &&& body) . filter ((>=2) . length) . init . subsequences) indices
>                            ++ concat [map (((i:) *** id) *** id) (subqexprs (fs!!i)) | i <- indices]
>     where indices = [0..length fs-1]
>           body = flatten . (QSym s) . keepIndices fs

> replace :: QExpr -> Location -> QExpr -> QExpr
> replace _ ([],[]) e' = e'
> replace (QUnSym s e) (0:is,children) e' = QUnSym s $ replace e (is,children) e'
> replace (QSym s es) (i:is,children) e' = QSym s (ys++replace e (is,children) e':xs)
>     where (ys,e:xs) = splitAt i es
> replace (QSym s es) ([],children) e' = flatten $ QSym s (e':elimIndices es children)


Las funciones que siguen son las que hacen el trabajo interesante:

Dada una expresión, una regla y una sustitución, rewriteWithSubst
encuentra todas las posibles reescrituras de la expresión original
aplicando la regla con una sustitución dada.

-- > rewriteWithSubst :: Expr -> Rule -> Subst Expr -> [Expr]
-- > rewriteWithSubst e rule sigma = [replace e' loc (unlift (rhs rule -/- sigma'))
-- >                                 | e' <- (unlifts . lift) e 
-- >                                 , l' <- unlifts (lhs rule -/- sigma')
-- >                                 , loc <- (map fst . filter ((==l') . snd)) (subexprs e')]
-- >     where sigma' = liftSubst sigma

Dada una expresión (la conclusión), una regla y una expresión inicial,
useRule devuelve la lista de posibles sustituciones que permitirían
aplicar una regla.

> useRule :: QExpr -> Rule -> QExpr -> [Subst QExpr]
> useRule e1 rule e2 = (map fst . filter ((e2==) . snd) . concat) 
>                      [map (id &&& rewrite rule e1 l) ss | (l,ss) <- matchings rule e1]


> rewrite :: Rule -> QExpr -> Location -> Subst QExpr -> QExpr
> rewrite r e l s = replace e l (rhs r -/- s)


> matchings ::  Rule -> QExpr -> [(Location, [Subst QExpr])]
> matchings rule = filter (not . null . snd) . map (id *** match (lhs rule)) . subqexprs

> t =  var "s" `equiv` var "q"  `equiv` (var "t" `disj` var "u")

-- > t' = var "q" `equiv` var "s" `equiv` (var "t" `disj` var "u")

-- > regla = rules (axioms!!1) !! 3

-- > t2 = var "s" `disj` (non (var "s"))
-- > t3 = var "s" `implies` var "t"

> printty :: Show a => [a] -> IO ()
> printty = mapM_ (putStrLn . U.toUTF8 . show) 


> e = dothis id "r /\\ s == (t => q) == p \\/ q == p == q"
> e' = dothis id "t => q == p /\\ q == r /\\ s"
> e2 = dothis id "p /\\ q == t => q == r /\\ s"
> dothis this = either (const $ error "") this . parseExp 